@@ -13,24 +13,15 @@
 package Directory::Queue::Set;
 use strict;
 use warnings;
-our $VERSION  = "1.10";
-our $REVISION = sprintf("%d.%02d", q$Revision: 1.12 $ =~ /(\d+)\.(\d+)/);
+our $VERSION  = "1.2";
+our $REVISION = sprintf("%d.%02d", q$Revision: 1.4 $ =~ /(\d+)\.(\d+)/);
 
 #
 # used modules
 #
 
-use No::Worries::Die qw(dief);
-
-#
-# return true if the given thing is a Directory::Queue object
-#
-
-sub _isdq ($) {
-    my($thing) = @_;
-
-    return(ref($thing) && $thing->isa("Directory::Queue"));
-}
+use Directory::Queue;
+use UNIVERSAL qw();
 
 #
 # object constructor
@@ -52,16 +43,17 @@ sub new : method {
 
 sub add : method {
     my($self, @list) = @_;
-    my($id);
+    my($dirq, $id);
 
-    foreach my $dirq (@list) {
-        dief("not a Directory::Queue object: %s", $dirq) unless _isdq($dirq);
-        $id = $dirq->id();
-        dief("duplicate queue in set: %s", $dirq->path())
-            if $self->{dirq}{$id};
-        $self->{dirq}{$id} = $dirq->copy();
+    foreach $dirq (@list) {
+	Directory::Queue::_fatal("not a Directory::Queue object: %s", $dirq)
+	    unless UNIVERSAL::isa($dirq, "Directory::Queue");
+	$id = $dirq->id();
+	Directory::Queue::_fatal("duplicate queue in set: %s", $dirq->path())
+	    if $self->{dirq}{$id};
+	$self->{dirq}{$id} = $dirq->copy();
     }
-    # reset our iterator
+    # reset the iterator
     delete($self->{elt});
 }
 
@@ -71,16 +63,17 @@ sub add : method {
 
 sub remove : method {
     my($self, @list) = @_;
-    my($id);
+    my($dirq, $id);
 
-    foreach my $dirq (@list) {
-        dief("not a Directory::Queue object: %s", $dirq) unless _isdq($dirq);
-        $id = $dirq->id();
-        dief("missing queue in set: %s", $dirq->path())
-            unless $self->{dirq}{$id};
-        delete($self->{dirq}{$id});
+    foreach $dirq (@list) {
+	Directory::Queue::_fatal("not a Directory::Queue object: %s", $dirq)
+	    unless UNIVERSAL::isa($dirq, "Directory::Queue");
+	$id = $dirq->id();
+	Directory::Queue::_fatal("missing queue in set: %s", $dirq->path())
+	    unless $self->{dirq}{$id};
+	delete($self->{dirq}{$id});
     }
-    # reset our iterator
+    # reset the iterator
     delete($self->{elt});
 }
 
@@ -88,20 +81,20 @@ sub remove : method {
 # get the next element of the queue set
 #
 
-sub next : method { ## no critic 'ProhibitBuiltinHomonyms'
+sub next : method {
     my($self) = @_;
-    my($name, $min_elt, $min_id);
+    my($id, $name, $min_elt, $min_id);
 
     return() unless $self->{elt};
-    foreach my $id (keys(%{ $self->{elt} })) {
-        $name = substr($self->{elt}{$id}, -14);
-        next if defined($min_elt) and $min_elt le $name;
-        $min_elt = $name;
-        $min_id = $id;
+    foreach $id (keys(%{ $self->{elt} })) {
+	$name = substr($self->{elt}{$id}, -14);
+	next if defined($min_elt) and $min_elt le $name;
+	$min_elt = $name;
+	$min_id = $id;
     }
     unless ($min_id) {
-        delete($self->{elt});
-        return();
+	delete($self->{elt});
+	return();
     }
     $min_elt = $self->{elt}{$min_id};
     $self->{elt}{$min_id} = $self->{dirq}{$min_id}->next();
@@ -115,12 +108,13 @@ sub next : method { ## no critic 'ProhibitBuiltinHomonyms'
 
 sub first : method {
     my($self) = @_;
+    my($id);
 
     return() unless $self->{dirq};
     delete($self->{elt});
-    foreach my $id (keys(%{ $self->{dirq} })) {
-        $self->{elt}{$id} = $self->{dirq}{$id}->first();
-        delete($self->{elt}{$id}) unless $self->{elt}{$id};
+    foreach $id (keys(%{ $self->{dirq} })) {
+	$self->{elt}{$id} = $self->{dirq}{$id}->first();
+	delete($self->{elt}{$id}) unless $self->{elt}{$id};
     }
     return($self->next());
 }
@@ -131,12 +125,12 @@ sub first : method {
 
 sub count : method {
     my($self) = @_;
-    my($count);
+    my($count, $id);
 
     return(0) unless $self->{dirq};
     $count = 0;
-    foreach my $id (keys(%{ $self->{dirq} })) {
-        $count += $self->{dirq}{$id}->count();
+    foreach $id (keys(%{ $self->{dirq} })) {
+	$count += $self->{dirq}{$id}->count();
     }
     return($count);
 }
@@ -170,10 +164,6 @@ This module can be used to put different queues into a set and browse
 them as one queue. The elements from all queues are merged together
 and sorted independently from the queue they belong to.
 
-This works both with L<Directory::Queue::Normal> and
-L<Directory::Queue::Simple> queues. Queues of different types can even
-be mixed.
-
 =head1 METHODS
 
 The following methods are available:
@@ -182,26 +172,29 @@ The following methods are available:
 
 =item new([DIRQ...])
 
-return a new Directory::Queue::Set object containing the given queue objects
-(class method)
+return a new Directory::Queue::Set object containing the given
+Directory::Queue objects (class method)
 
 =item add([DIRQ...])
 
-add the given queue objects to the queue set; resetting the iterator
+add the given Directory::Queue objects to the queue set;
+resetting the iterator
 
 =item remove([DIRQ...])
 
-remove the given queue objects from the queue set; resetting the iterator
+remove the given Directory::Queue objects from the queue set;
+resetting the iterator
 
 =item first()
 
-return the first (queue, element) couple in the queue set, resetting the
-iterator; return an empty list if the queue is empty
+return the first (queue, element) couple in the queue set,
+resetting the iterator;
+return an empty list if the queue is empty
 
 =item next()
 
-return the next (queue, element) couple in the queue set; return an empty
-list if there is no next element
+return the next (queue, element) couple in the queue set; return an
+empty list if there is no next element
 
 =item count()
 
@@ -209,12 +202,8 @@ return the total number of elements in all the queues of the set
 
 =back
 
-=head1 SEE ALSO
-
-L<Directory::Queue>.
-
 =head1 AUTHOR
 
 Lionel Cons L<http://cern.ch/lionel.cons>
 
-Copyright (C) CERN 2010-2018
+Copyright CERN 2010
